;; Generated package description from nameless.el  -*- no-byte-compile: t -*-
(define-package "nameless" "1.0.2.0.20230112.95905" "Hide package namespace in your emacs-lisp code" '((emacs "24.4")) :commit "e468f3eea4518b9827419611868c897dce20453f" :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :keywords '("convenience" "lisp") :url "https://github.com/Malabarba/nameless")
